/*
 * Copyright (C) 2002, 2003 Jan Panteltje <panteltje@yahoo.com>,
 *
 * Modified by Zachary Brewster-Geisz, 2003, to work on big-endian
 * machines.
 *
 * Modified by Henry Mason, 2003, to use both PNG and BMP, and to use
 * the dvdauthor submux format.
 *
 * Modified and copy right Jan Panteltje 2002
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * With many changes by Scott Smith (trckjunky@users.sourceforge.net)
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#include <netinet/in.h>

#define FALSE 0
#define TRUE (!FALSE)

#define CBUFSIZE 65536
#define PSBUFSIZE 10

static unsigned int add_offset = 450;
static unsigned int ofs, ofs1, pts, spts, subi, subs, subno;
static unsigned char next_bits, sub[65536];
static int debug, have_bits, forced;
static char *output;

typedef struct
{
  unsigned char r, g, b, t;
} palt;

static palt bpal[16] =
{
  {   0,   0,   0, 0 },
  { 127,   0,   0, 0 },
  {   0, 127,   0, 0 },
  { 127, 127,   0, 0 },
  {   0,   0, 127, 0 },
  { 127,   0, 127, 0 },
  {   0, 127, 127, 0 },
  { 127, 127, 127, 0 },
  { 192, 192, 192, 0 },
  { 128,   0,   0, 0 },
  {   0, 128,   0, 0 },
  { 128, 128,   0, 0 },
  {   0,   0, 128, 0 },
  { 128,   0, 128, 0 },
  {   0, 128, 128, 0 },
  { 128, 128, 128, 0 }
};

struct spu
{
  unsigned char *img;
  unsigned int x0, y0, xd, yd, pts[2], subno, force_display, nummap;
  struct colormap *map;
  struct spu *next;
};

static struct spu *spus = 0;

struct button
{
  char *name;
  int autoaction;
  int x1, y1, x2, y2;
  char *up, *down, *left, *right;
  int grp;
};

struct dispdetails
{
  int pts[2];
  int numpal;
  u_int32_t palette[16];
  int numcoli;
  u_int32_t coli[6];
  int numbuttons;
  struct button *buttons;
  struct dispdetails *next;
};

static struct dispdetails *dd = 0;

struct colormap
{
  u_int16_t color;
  u_int16_t contrast;
  int x1, y1, x2, y2;
};

static inline int RGBclamp (int x)
{
  if (x < 0)
    return 0;
  if (x > 255)
    return 255;
  return x;
}

static inline int YCrCb2R(int Y,int Cr,int Cb)
{
  return RGBclamp ((500 + 1164 * (Y - 16) + 1596 * (Cr - 128)) / 1000);
}

static inline int YCrCb2G(int Y,int Cr,int Cb)
{
  return RGBclamp ((500 + 1164 * (Y - 16) - 813 * (Cr - 128) - 391 * (Cb - 128)) / 1000);
}

static inline int YCrCb2B(int Y,int Cr,int Cb)
{
  return RGBclamp ((500 + 1164 * (Y - 16) + 2018 * (Cb - 128)) / 1000);
}
static unsigned int
read4 (unsigned char *p)
{
  return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
}

static unsigned int
read2 (unsigned char *p)
{
  return (p[0] << 8) | p[1];
}

static char *
readpstr (char *b, int *i)
{
  char *s = strdup (b + i[0]);
  i[0] += strlen (s) + 1;
  return s;
}

static unsigned char
get_next_bits ()
{
  if (!have_bits)
  {
    next_bits = sub[ofs++];
    have_bits = TRUE;
    return next_bits >> 4;
  }
  have_bits = FALSE;
  return next_bits & 15;
}

static unsigned int
getpts (unsigned char *buf)
{
  if (!(buf[1] & 0xc0) || (buf[2] < 4) ||
      ((buf[3] & 0xe1) != 0x21) ||
      ((buf[5] & 1) != 1) || ((buf[7] & 1) != 1))
    return -1;

  return (buf[7] >> 1) +
    ((unsigned int) buf[6] << 7) +
    (((unsigned int) buf[5] & 254) << 14) +
    ((unsigned int) buf[4] << 22) +
    (((unsigned int) buf[3] & 14) << 29);
}

static void
addspu (struct spu *s)
{
  struct spu **f = &spus;

  while (*f)
    f = &(f[0]->next);

  *f = s;
}

static void
adddd (struct dispdetails *d)
{
  struct dispdetails **dp = &dd;

  while (*dp)
    dp = &(dp[0]->next);

  *dp = d;
}

static int
dvddecode ()
{
  unsigned int io;
  unsigned short int size, dsize, i, x, y, t;
  unsigned char c;
  struct spu *s;

  size = (((unsigned int) sub[0]) << 8) + sub[1];
  dsize = (((unsigned int) sub[2]) << 8) + sub[3];

  if (debug > 1)
    fprintf (stderr, "packet: 0x%x bytes, 0x%x bytes data\n", size, dsize);

  s = malloc (sizeof (struct spu));
  memset (s, 0, sizeof (struct spu));

  s->subno = subno++;

  s->pts[0] = s->pts[1] = -1;
  s->nummap = 1;
  s->map = malloc (sizeof (struct colormap));
  memset (s->map, 0, sizeof (struct colormap));
  s->map[0].x2 = 0x7ffffff;
  s->map[0].y2 = 0x7ffffff;
  i = dsize + 4;

  t = sub[dsize] * 256 + sub[dsize + 1];

  if (debug > 2)
    fprintf (stderr, "time offset: %d sub[%d]=0x%02x sub[%d]=0x%02x\n",
             t, dsize, sub[dsize], dsize + 1, sub[dsize + 1]);

  while (i < size)
  {
    c = sub[i];
    if (debug > 4)
      fprintf (stderr, "i: %d cmd: 0x%x\n", i, c);

    switch (c)
    {
      case 0x0: //force start display
        i++;
        s->force_display = TRUE;
        s->pts[0] = t * 1024 + spts;
        if (debug > 4)
          fprintf (stderr, "cmd: force start display\n");
        break;

      case 0x01:
        i++;
        s->pts[0] = t * 1024 + spts;
        if (debug > 4)
          fprintf (stderr, "cmd: start display\n");
        break;

      case 0x02:
        if (debug > 4)
          fprintf (stderr, "cmd: end display\n");
        s->pts[1] = t * 1024 + spts;
        i++;
        break;

      case 0x03:
        if (debug > 4)
          fprintf (stderr, "cmd: palette %02x%02x\n", sub[i + 1], sub[i + 2]);

        s->map[0].color = read2 (sub + i + 1);
        i += 3;
        break;

      case 0x04:
        if (debug > 4)
          fprintf (stderr, "cmd: t-palette %02x%02x\n", sub[i + 1], sub[i + 2]);

        s->map[0].contrast = read2 (sub + i + 1);
        i += 3;
        break;

      case 0x05:
        s->x0 = ((((unsigned int) sub[i + 1]) << 4) + (sub[i + 2] >> 4));
        s->xd = (((sub[i + 2] & 0x0f) << 8) + sub[i + 3]) - s->x0 + 1;

        s->y0 = ((((unsigned int) sub[i + 4]) << 4) + (sub[i + 5] >> 4));
        s->yd = (((sub[i + 5] & 0x0f) << 8) + sub[i + 6]) - s->y0 + 1;

        if (debug > 4)
          fprintf (stderr, "cmd: img ofs %d,%d  size: %d,%d\n", s->x0,
                   s->y0, s->xd, s->yd);
        i += 7;
        break;

      case 0x06:
        ofs = (((unsigned int) sub[i + 1]) << 8) + sub[i + 2];
        ofs1 = (((unsigned int) sub[i + 3]) << 8) + sub[i + 4];
        if (debug > 4)
          fprintf (stderr, "cmd: image offsets 0x%x 0x%x\n", ofs, ofs1);
        i += 5;
        break;

      case 0xff:
        if (debug > 4)
          fprintf (stderr, "cmd: end cmd\n");

        if (i + 5 > size)
        {
          if (debug > 4)
            fprintf (stderr,
                     "short end command i=%d size=%d (%d bytes)\n",
                     i, size, size - i);

          i = size;
          break;
        }

        t = sub[i + 1] * 256 + sub[i + 2];
        if (debug > 4)
        {
          fprintf (stderr, "next packet time: %d  end: 0x%02x%02x\n",
                   t, sub[i + 3], sub[i + 4]);
        }

        if ((sub[i + 3] != sub[dsize + 2]) || (sub[i + 4] != sub[dsize + 3]))
        {
          if (debug > 0)
            fprintf (stderr, "invalid control header (%02x%02x != %02x%02x) dsize=%d!\n",
                     sub[i + 3], sub[i + 4], sub[dsize + 2], sub[dsize + 3], dsize);

          if (debug > 3)
            fprintf (stderr, "i: %d\n", i);

          i = size;
          break;
        }

        i += 5;
        break;

      default:
        if (debug > 0)
          fprintf (stderr, "invalid sequence in control header (%02x)!\n", c);
        return -1;
    }
  }

  have_bits = FALSE;
  x = y = 0;
  io = 0;
  s->img = malloc (s->xd * s->yd);

  while ((ofs < dsize) && (y < s->yd))
  {
    i = get_next_bits ();

    if (i < 4)
    {
      i = (i << 4) + get_next_bits ();
      if (i < 16)
      {
        i = (i << 4) + get_next_bits ();
        if (i < 0x40)
        {
          i = (i << 4) + get_next_bits ();
          if (i < 256)
          {
            have_bits = FALSE;

            y += 2;
            while (x++ != s->xd)
              s->img[io++] = i;
            x = 0;
            if ((y >= s->yd) && !(y & 1))
            {
              y = 1;
              io = s->xd;
              ofs = ofs1;
            }
            else
              io += s->xd;
            continue;
          }
        }
      }
    }

    c = i & 3;
    i = i >> 2;
    while (i--)
    {
      s->img[io++] = c;
      if (++x == s->xd)
      {
        y += 2;
        x = 0;
        if ((y >= s->yd) && !(y & 1))
        {
          y = 1;
          io = s->xd;
          ofs = ofs1;
        }
        else
          io += s->xd;
        have_bits = FALSE;
      }
    }
  }

  if (s->pts[0] == -1)
    return 0;

  s->pts[0] += add_offset;
  if (s->pts[1] != -1)
    s->pts[1] += add_offset;

  addspu (s);

  if (debug > 2)
    fprintf (stderr, "ofs: 0x%x y: %d\n", ofs, y);

  return 0;
}

static void
ycrcb_to_rgb (int *Y, int *Cr, int *Cb)
{
  int R, G, B;
  R = YCrCb2R (*Y, *Cr, *Cb);
  G = YCrCb2G (*Y, *Cr, *Cb);
  B = YCrCb2B (*Y, *Cr, *Cb);
  *Y = R;
  *Cr = G;
  *Cb = B;
}

static void
absorb_palette (struct dispdetails *d)
{
  int i;
  for (i = 0; i < d->numpal; i++)
  {
    int Y, Cr, Cb;

    Y = (d->palette[i] >> 16) & 255;
    Cr = (d->palette[i] >> 8) & 255;
    Cb = (d->palette[i]) & 255;
    bpal[i].r = YCrCb2R (Y, Cr, Cb);
    bpal[i].g = YCrCb2G (Y, Cr, Cb);
    bpal[i].b = YCrCb2B (Y, Cr, Cb);
  }
}

static void
pluck_dd ()
{
  struct dispdetails *d = dd;
  int i;

  dd = d->next;
  absorb_palette (d);
  for (i = 0; i < d->numbuttons; i++)
  {
    free (d->buttons[i].name);
    free (d->buttons[i].up);
    free (d->buttons[i].down);
    free (d->buttons[i].left);
    free (d->buttons[i].right);
  }
  free (d->buttons);
  free (d);
}

static unsigned char
cmap_find (int x, int y, struct colormap *map, int nummap, int ci)
{
  int i;
  unsigned char cix = 0;

  for (i = 0; i < nummap; i++)
    if (x >= map[i].x1 && y >= map[i].y1 && x <= map[i].x2 && y <= map[i].y2)
      cix = (((map[i].contrast >> (ci * 4)) & 15) << 4) |
        (((map[i].color >> (ci * 4)) & 15));
  return cix;
}

static int
write_ppm (char *file_name, struct spu *s, struct colormap *map, int nummap)
{
  FILE *f;
  unsigned int x, y;
  unsigned char cix;

  f = fopen (file_name, "w");

  fprintf (f, "P5\n" "%d %d\n" "255\n", s->xd, s->yd);

  for (y = 0; y < s->yd; y++)
  {
    for (x = 0; x < s->xd; x++)
    {
      int res;

      cix = cmap_find (x + s->x0, y + s->y0, map, nummap, s->img[y * s->xd + x]);

      if (cix >> 4)
        res = (bpal[cix & 15].r + bpal[cix & 15].g + bpal[cix & 15].b) / 3 * ((cix >> 4) * 17);
      else
        res = 0;

      putc ((255 - res) & 0xff, f);
    }
  }

  putc ('\n', f);

  fclose (f);

  return 0;
}

static void
write_pts (FILE *srtx_file, char *preamble, int pts)
{
  int temp, h, m, s, ms;

  temp = pts;
  temp /= 90;
  h = temp / 3600000;
  temp %= 3600000;
  m = temp / 60000;
  temp %= 60000;
  s = temp / 1000;
  temp %= 1000;
  ms = temp;

  fprintf (srtx_file, "%02d:%02d:%02d,%03d", h, m, s, ms);
}

static void
write_spu (FILE *srtx_file, struct spu *s, struct dispdetails *d)
{
  char buf[256];

  if (d)
    absorb_palette (d);

  if (!forced || s->force_display)
  {
    fprintf (srtx_file, "%d\n", s->subno + 1);

    sprintf (buf, "%s%04d.pgm", output, s->subno + 1);
    write_ppm (buf, s, s->map, s->nummap);

    write_pts (srtx_file, "start", s->pts[0]);
    if (s->pts[1] != -1)
    {
      fprintf (srtx_file, " --> ");
      write_pts (srtx_file, "end", s->pts[1]);
    }

    fprintf (srtx_file, "\n");
    fprintf (srtx_file, "%s.txt\n\n", buf);
  }
}

static void
flushspus (FILE *srtx_file, unsigned int lasttime)
{
  while (spus)
  {
    struct spu *s = spus;
    if (s->pts[0] >= lasttime)
      return;
    spus = spus->next;

    while (dd && dd->pts[1] < s->pts[0] && dd->pts[1] != -1)
      pluck_dd ();

    if (dd && (dd->pts[0] < s->pts[1] || s->pts[1] == -1) &&
        (dd->pts[1] > s->pts[0] || dd->pts[1] == -1))
      write_spu (srtx_file, s, dd);
    else
      write_spu (srtx_file, s, 0);
    if (s->img)
      free (s->img);
    if (s->map)
      free (s->map);
    free (s);
  }
}

static void
process_sub (FILE *sub_file, FILE *srtx_file, int sid)
{
  unsigned short int package_length;
  unsigned char cbuf[CBUFSIZE];
  unsigned char psbuf[PSBUFSIZE];
  unsigned int c, next_word;
  int firstvideo = -1;

  while (fread (&c, sizeof (c), 1, sub_file) == 1)
  {
    c = ntohl (c);
    if (c == 0x000001ba)
    { // start PS (Program stream)
      static unsigned int old_system_time = -1;
      unsigned int new_system_time;
    l_01ba:
      if (debug > 5)
        fprintf (stderr, "pack_start_code\n");

      if (fread (psbuf, sizeof (char), PSBUFSIZE, sub_file) < 1)
        break;

      new_system_time = (psbuf[4] >> 3) + (psbuf[3] * 32) +
        ((psbuf[2] & 3) * 32 * 256) +
        ((psbuf[2] & 0xf8) * 32 * 128) +
        (psbuf[1] * 1024 * 1024) +
        ((psbuf[0] & 3) * 1024 * 1024 * 256) +
        ((psbuf[0] & 0x38) * 1024 * 1024 * 256 * 2);

      if (new_system_time < old_system_time)
      {
        if (old_system_time != -1)
        {
          if (debug > 0)
            printf
              ("Time changed in stream header, use old time as offset for timecode in subtitle stream\n");
          add_offset += old_system_time;
        }
      }
      old_system_time = new_system_time;

      flushspus (srtx_file, old_system_time);

      if (debug > 5)
      {
        fprintf (stderr, "system time: %d\n", new_system_time);
      }
    }
    else if (c == 0x1b9)
    {
      if (debug > 5)
        fprintf (stderr, "end packet\n");
    }
    else
    {
      fread (&package_length, 2, 1, sub_file);
      package_length = ntohs (package_length);
      if (package_length != 0)
      {

        switch (c)
        {
        case 0x01bb:
          if (debug > 5)
            fprintf (stderr, "system header\n");
          break;
        case 0x01bf:
          if (debug > 5)
            fprintf (stderr, "private stream 2\n");
          break;
        case 0x01bd:
          if (debug > 5)
            fprintf (stderr, "private stream\n");

          fread (cbuf, 1, package_length, sub_file);

          next_word = getpts (cbuf);
          if (next_word != -1)
          {
            pts = next_word;
          }

          next_word = cbuf[2] + 3;

          if (debug > 5)
            for (c = 0; c < next_word; c++)
              fprintf (stderr, "0x%02x ", cbuf[c]);

          if (debug > 5)
            fprintf (stderr, "tid: %d\n", pts);

          if ((cbuf[next_word] == 0x70))
            fprintf (stderr, "substr: %d\n", cbuf[next_word + 1]);

          if (cbuf[next_word] == sid + 32)
          {
            if ((debug < 6) && (debug > 1))
            {
              fprintf (stderr,
                       "id: 0x%x 0x%x %d  tid: %d\n",
                       cbuf[next_word], package_length, next_word, pts);
            }

            if (!subi)
            {
              subs =
                ((unsigned int) cbuf[next_word + 1] <<
                 8) + cbuf[next_word + 2];

              spts = pts;
            }

            memcpy (sub + subi, cbuf + next_word + 1,
                    package_length - next_word - 1);

            if (debug > 1)
            {
              fprintf (stderr, "found %d bytes of data\n",
                       package_length - next_word - 1);
            }

            subi += package_length - next_word - 1;

            if (debug > 2)
            {
              fprintf (stderr,
                       "subi: %d (0x%x)  subs: %d (0x%x) b-a-1: %d (0x%x)\n",
                       subi, subi, subs, subs,
                       package_length - next_word - 1,
                       package_length - next_word - 1);
            }

            if (subs == subi)
            {
              subi = 0;

              next_word = dvddecode ();

              if (next_word)
              {
                fprintf (stderr,
                         "found unreadable subtitle at %.2fs, skipping\n",
                         (double) spts / 90000);
                continue;
              }
            }
          }
          package_length = 0;
          break;
        case 0x01e0:
          if (firstvideo == -1)
          {
            fread (cbuf, 1, package_length, sub_file);
            firstvideo = getpts (cbuf);
            add_offset -= firstvideo;
            package_length = 0;
          }
          if (debug > 5)
            fprintf (stderr, "video stream 0\n");
          break;
        case 0x01e1:
        case 0x01e2:
        case 0x01e3:
        case 0x01e4:
        case 0x01e5:
        case 0x01e6:
        case 0x01e7:
        case 0x01e8:
        case 0x01e9:
        case 0x01ea:
        case 0x01eb:
        case 0x01ec:
        case 0x01ed:
        case 0x01ee:
        case 0x01ef:
          if (debug > 5)
            fprintf (stderr, "video stream %d\n", c - 0x1e0);
          break;
        case 0x01be:
          if (debug > 5)
            fprintf (stderr, "padding stream %d bytes\n", package_length);
          fread (cbuf, 1, package_length, sub_file);
          if (package_length > 30)
          {
            int i;

            package_length = 0;
            i = 0;
            if (strcmp ((char *) cbuf + i, "dvdauthor-data"))
              break;
            i = 15;
            if (cbuf[i] != 1)
              break;
            switch (cbuf[i + 1])
            {
            case 1:
              { // subtitle/menu color and button information
                struct dispdetails *d;
                i += 3;
                d = malloc (sizeof (struct dispdetails));
                memset (d, 0, sizeof (struct dispdetails));
                d->pts[0] = read4 (cbuf + i);
                d->pts[1] = read4 (cbuf + i + 4);
                i += 8;
                while (cbuf[i] != 0xff)
                {
                  switch (cbuf[i])
                  {
                  case 1:
                    {
                      int j;

                      d->numpal = 0;
                      for (j = 0; j < cbuf[i + 1]; j++)
                      {
                        int c =
                          (cbuf[i + 2 + 3 * j] << 16) | (cbuf[i + 3 + 3 * j] << 8) | (cbuf[i + 4 + 3 * j]);
                        d->palette[j] = c;
                        d->numpal++;
                      }
                      i += 2 + 3 * d->numpal;
                      break;
                    }
                  case 2:
                    {
                      int j;

                      d->numcoli = cbuf[i + 1];
                      for (j = 0; j < 2 * d->numcoli; j++)
                        d->coli[j] = read4 (cbuf + i + 2 + j * 4);
                      i += 2 + 8 * d->numcoli;
                      break;
                    }
                  case 3:
                    {
                      int j;
                      d->numbuttons = cbuf[i + 1];
                      d->buttons =
                        malloc (d->numbuttons * sizeof (struct button));
                      for (j = 0; j < d->numbuttons; j++)
                      {
                        struct button *b = &d->buttons[j];
                        b->name = readpstr ((char *) cbuf, &i);
                        i += 2;
                        b->autoaction = cbuf[i++];
                        if (!b->autoaction)
                        {
                          b->grp = cbuf[i];
                          b->x1 = read2 (cbuf + i + 1);
                          b->y1 = read2 (cbuf + i + 3);
                          b->x2 = read2 (cbuf + i + 5);
                          b->y2 = read2 (cbuf + i + 7);
                          i += 9;
                          // up down left right
                          b->up = readpstr ((char *) cbuf, &i);
                          b->down = readpstr ((char *) cbuf, &i);
                          b->left = readpstr ((char *) cbuf, &i);
                          b->right = readpstr ((char *) cbuf, &i);
                        }
                      }
                      break;
                    }

                  default:
                    fprintf (stderr,
                             "ERR:  unknown dvd info packet command: %d\n",
                             cbuf[i]);
                    exit (1);
                  }
                }
                adddd (d);
                break;
              }
            }
          }
          package_length = 0;
          break;
        case 0x01c0:
        case 0x01c1:
        case 0x01c2:
        case 0x01c3:
        case 0x01c4:
        case 0x01c5:
        case 0x01c6:
        case 0x01c7:
        case 0x01c8:
        case 0x01c9:
        case 0x01ca:
        case 0x01cb:
        case 0x01cc:
        case 0x01cd:
        case 0x01ce:
        case 0x01cf:
        case 0x01d0:
        case 0x01d1:
        case 0x01d2:
        case 0x01d3:
        case 0x01d4:
        case 0x01d5:
        case 0x01d6:
        case 0x01d7:
        case 0x01d8:
        case 0x01d9:
        case 0x01da:
        case 0x01db:
        case 0x01dc:
        case 0x01dd:
        case 0x01de:
        case 0x01df:
          if (debug > 5)
            fprintf (stderr, "audio stream %d\n", c - 0x1c0);
          break;
        default:
          if (debug > 0)
            fprintf (stderr, "unknown header %x\n", c);
          next_word = (c << 16) | package_length;
          package_length = 2;
          while (next_word != 0x1ba)
          {
            next_word = next_word << 8;
            if (fread (&next_word, 1, 1, sub_file) < 1)
              break;
            package_length++;
          }

          if (debug > 0)
            fprintf (stderr,
                     "skipped %d bytes of garbage\n", package_length);
          goto l_01ba;
        }
        fread (cbuf, 1, package_length, sub_file);
      }
    }
  }
}

#define bps(n,R,G,B) do { bpal[n].r=R; bpal[n].g=G; bpal[n].b=B; } while (0)

static void
load_palette (const unsigned char *palet_file)
{
  FILE *file;
  int n, rgb = FALSE;
  char *temp;

  temp = strrchr ((char *) palet_file, '.');
  if (temp != NULL)
  {
    if (strcmp (temp, ".rgb") == 0)
      rgb = TRUE;
  }

  file = fopen ((char *) palet_file, "r");
  if (file != NULL)
  {
    for (n = 0; n < 16; n++)
    {
      int r, g, b;
      fscanf (file, "%02x%02x%02x", &r, &g, &b);
      if (!rgb)
        ycrcb_to_rgb (&r, &g, &b);
      bpal[n].r = r;
      bpal[n].g = g;
      bpal[n].b = b;

      if (debug > 3)
        fprintf (stderr, "pal: %d #%02x%02x%02x\n", n,
                 bpal[n].r, bpal[n].g, bpal[n].b);

    }
    fclose (file);
  }
  else
    fprintf (stderr, "unable to open %s, using defaults\n", palet_file);
}

static void
usage (const char *name)
{
  fprintf (stdout, "Usage:\n");
  fprintf (stdout, "  %s [OPTION...] <vobsub basename>\n", name);
  fprintf (stdout, "\n");
  fprintf (stdout, "Help Options:\n");
  fprintf (stdout, "  -h, --help                  Show help options\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "Application Options:\n");
  fprintf (stdout, "  -o, --output=<filename>     Use filename for output (default: stdout)\n");
  fprintf (stderr, "  -s, --sid=<sid>             The subtitle id (default: 0)\n");
  fprintf (stderr, "  -f, --forced                Extract only forced subtitles\n");
  fprintf (stderr, "  -v, --verbose               Increase verbosity level\n");
  fprintf (stdout, "\n");
}

static const char *shortopts = "hfvs:o:p:";
static const struct option longopts[] =
{
  { "help",    no_argument,       NULL, 'h' },
  { "forced",  no_argument,       NULL, 'f' },
  { "verbose", no_argument,       NULL, 'v' },
  { "sid",     required_argument, NULL, 's' },
  { "output",  required_argument, NULL, 'o' },
  { "palette", required_argument, NULL, 'p' },
  { NULL,      0,                 NULL,  0  }
};

int
main (int argc, char **argv)
{
  FILE *sub_file, *srtx_file;

  unsigned int sid = 0;
  unsigned char nbuf[FILENAME_MAX], *palet_file = NULL;
  int option;

  while ((option = getopt_long (argc, argv, shortopts, longopts, NULL)) != EOF)
  {
    switch (option)
    {
      case 's':
        sid = atoi (optarg);
        break;
      case 'o':
        output = optarg;
        break;
      case 'p':
        palet_file = (unsigned char *) optarg;
        break;
      case 'f':
        forced ++;
        break;
      case 'v':
        debug ++;
        break;
      case 'h':
        usage (argv[0]);
        return EXIT_SUCCESS;
        break;
      default:
        fprintf (stderr, "Unknown option. Use -h to list all valid options.\n");
        return EXIT_FAILURE;
        break;
    }
  }

  if (optind >= argc)
  {
    usage (argv[0]);
    return EXIT_FAILURE;
  }

  if (!output)
    output = argv[optind];

  if (palet_file)
    load_palette (palet_file);

  if (strlen (output) > FILENAME_MAX)
  {
    fprintf (stderr, "Error max length of base for filename creation is %d characters\n", FILENAME_MAX);
    return -1;
  }

  sprintf ((char *) nbuf, "%s.sub", argv[optind]);
  sub_file = fopen ((char *) nbuf, "r");
  if (!sub_file)
  {
    fprintf (stderr, "Error opening file %s\n", argv[optind]);
    exit (-1);
  }

  sprintf ((char *) nbuf, "%s.srtx", output);
  srtx_file = fopen ((char *) nbuf, "w+");
  if (!srtx_file)
  {
    fclose (sub_file);
    fprintf (stderr, "Error opening file %s\n", nbuf);
    exit (-1);
  }

  process_sub (sub_file, srtx_file, sid);

  fclose (sub_file);

  flushspus (srtx_file, 0x7fffffff);
  fclose (srtx_file);

  fprintf (stdout, "%u files generated\n", subno);

  return 0;
}

